"""
Core class for micromind. Supports helper function for exports. Out-of-the-box
multi-gpu and FP16 training with HF Accelerate and much more.

Authors:
    - Francesco Paissan, 2023
"""
from typing import Dict, Union, Tuple, Callable, List
from abc import ABC, abstractmethod
from dataclasses import dataclass
from argparse import Namespace
from pathlib import Path
from loguru import logger
from tqdm import tqdm
import shutil

from accelerate import Accelerator
import torch
import os

from .utils.helpers import select_and_load_checkpoint, get_random_string
from .utils.checkpointer import Checkpointer

# This is used ONLY if you are not using argparse to get the hparams
default_cfg = {
    "output_folder": "results",
    "experiment_name": "micromind_exp",
    "opt": "adam",  # this is ignored if you are overriding the configure_optimizers
    "lr": 0.001,  # this is ignored if you are overriding the configure_optimizers
    "debug": False,
}


@dataclass
class Stage:
    """enum to track training stage"""

    train: int = 0
    val: int = 1
    test: int = 2


class Metric:
    """
    Class for tracking evaluation metrics during training.

    This class allows you to create custom evaluation metrics by providing a
    function to compute the metric and specifying a reduction method.

    Arguments
    ---------
        name : str
            The name of the metric.
        fn : Callable
            A function that computes the metric given predictions and batch data.
        reduction : Optional[str]
            The reduction method for the metric ('sum' or 'mean'). Default is 'mean'.

    Returns
    -------
        Reduced metric. Optionally, you can access the metric history
        before call reduce(clear=True) : torch.Tensor

    Example
    -------
    .. doctest::

        >>> from micromind import Metric, Stage
        >>> import torch

        >>> def custom_metric(pred, batch):
        ...     # Replace this with your custom metric calculation
        ...     return pred - batch

        >>> metric = Metric("Custom Metric", custom_metric, reduction="mean")
        >>> pred = torch.tensor([1.0, 2.0, 3.0])
        >>> batch = torch.tensor([0.5, 1.5, 2.5])
        >>> metric(pred, batch, stage=Stage.train)
        >>> metric.history
        {0: [tensor([0.5000, 0.5000, 0.5000])], 1: [], 2: []}
        >>> metric.reduce(Stage.train)
        0.5
    """

    def __init__(self, name: str, fn: Callable, reduction="mean"):
        self.name = name
        self.fn = fn
        self.reduction = reduction
        self.history = {s: [] for s in [Stage.train, Stage.val, Stage.test]}

    def __call__(self, pred, batch, stage, device="cpu"):
        if pred.device != device:
            pred = pred.to(device)
        dat = self.fn(pred, batch)
        if dat.ndim == 0:
            dat = dat.unsqueeze(0)

        self.history[stage].append(self.fn(pred, batch))

    def reduce(self, stage, clear=False):
        """
        Compute and return the metric for a given prediction and batch data.

        Arguments
        -------
            pred : torch.Tensor
                The model's prediction.
            batch : torch.Tensor
                The ground truth or target values.
            stage : Stage
                The current stage (e.g., Stage.train).
            device Optional[str]
                The device on which to perform the computation. Default is 'cpu'.
        """

        if self.reduction == "mean":
            if clear or (
                self.history[stage][-1].shape[0] != self.history[stage][0].shape[0]
            ):
                tmp = torch.stack(self.history[stage][:-1]).mean()
            else:
                tmp = torch.stack(self.history[stage]).mean()
        elif self.reduction == "sum":
            if (
                clear
                or self.history[stage][-1].shape[0] != self.history[stage][0].shape[0]
            ):
                tmp = torch.stack(self.history[stage][:-1]).sum()
            else:
                tmp = torch.stack(self.history[stage]).sum()

        if clear:
            self.history[stage] = []
        return tmp.item()


class MicroMind(ABC):
    """
    MicroMind is an abstract base class for creating and training deep learning
    models. Handles training on multi-gpu via accelerate (using DDP and other
    distributed training strategies). It automatically handles the device
    management for the training and the micromind's export capabilities to onnx,
    OpenVino and TFLite.

    Arguments
    ---------
        hparams : Optional[Namespace]
            Hyperparameters for the model. Default is None.

    """

    def __init__(self, hparams=None):
        if hparams is None:
            hparams = Namespace(**default_cfg)

        # here we should handle devices etc.
        self.modules = torch.nn.ModuleDict({})  # init empty modules dict
        self.hparams = hparams
        self.input_shape = None

        self.device = "cpu"  # used just to init the models
        self.accelerator = Accelerator()

    @abstractmethod
    def forward(self, batch):
        """
        Forward step of the class. It gets called during inference and optimization.
        This method should be overwritten for specific applications.

        Arguments
        ---------
            batch : torch.Tensor
                Batch as output from the defined DataLoader.

        Returns
        -------
            pred : Union[torch.Tensor, Tuple]
                Predictions - this depends on the task.
        """
        pass

    @abstractmethod
    def compute_loss(self, pred, batch):
        """
        Computes the cost function for the optimization process.  It return a
        tensor on which backward() is called. This method should be overwritten
        for the specific application.

        Arguments
        ---------
            pred : Union[torch.Tensor, Tuple]
                Output of the forward() function
            batch : torch.Tensor
                Batch as defined from the DataLoader.

        Returns
        -------
            loss : torch.Tensor
                Compute cost function.
        """
        pass

    def set_input_shape(self, input_shape: Tuple = (3, 224, 224)):
        """Setter function for input_shape.

        Arguments
        ---------
            input_shape : Tuple
                Input shape of the forward step.

        """
        self.input_shape = input_shape

    def load_modules(self, checkpoint_path: Union[Path, str]):
        """Loads models for path.

        Arguments
        ---------
            checkpoint_path : Union[Path, str]
                Path to the checkpoint where the modules are stored.

        """
        dat = torch.load(checkpoint_path)

        modules_keys = list(self.modules.keys())
        for k in self.modules:
            self.modules[k].load_state_dict(dat[k])

            modules_keys.remove(k)

        if len(modules_keys) != 0:
            print(modules_keys)
            breakpoint()
            logger.info(f"Couldn't find a state_dict for modules {modules_keys}.")

    def export(
        self, save_dir: Union[Path, str], out_format: str = "onnx", input_shape=None
    ) -> None:
        """
        Export the model to a specified format for deployment.
        TFLite and OpenVINO need a Linux machine to be exported.


        Arguments
        ---------
        save_dir : Union[Path, str]
            The directory where the exported model will be saved.
        out_format : Optional[str]
            The format for exporting the model. Default is 'onnx'.
        input_shape : Optional[Tuple]
            The input shape of the model. If not provided, the input shape
            specified during model creation is used.

        """
        from micromind import convert

        if not isinstance(save_dir, Path):
            save_dir = Path(save_dir)
        save_dir = save_dir.joinpath(self.hparams.experiment_name)

        self.set_input_shape(input_shape)
        assert (
            self.input_shape is not None
        ), "You should pass the input_shape of the model."

        if out_format == "onnx":
            convert.convert_to_onnx(
                self, save_dir.joinpath("model.onnx"), replace_forward=True
            )
        elif out_format == "openvino":
            convert.convert_to_openvino(self, save_dir, replace_forward=True)
        elif out_format == "tflite":
            convert.convert_to_tflite(self, save_dir, replace_forward=True)

    def configure_optimizers(self):
        """Configures and defines the optimizer for the task. Defaults to adam
        with lr=0.001; It can be overwritten by either passing arguments from the
        command line, or by overwriting this entire method.

        Returns
        ---------
           Optimizer and learning rate scheduler
           (not implemented yet). : Tuple[torch.optim.Adam, None]

        """
        assert self.hparams.opt in [
            "adam",
            "sgd",
        ], f"Optimizer {self.hparams.opt} not supported."
        if self.hparams.opt == "adam":
            opt = torch.optim.Adam(self.modules.parameters(), self.hparams.lr)
        elif self.hparams.opt == "sgd":
            opt = torch.optim.SGD(self.modules.parameters(), self.hparams.lr)
        return opt, None  # None is for learning rate sched

    def __call__(self, *x, **xv):
        """Just forwards everything to the forward method."""
        return self.forward(*x, **xv)

    def on_train_start(self):
        """Initializes the optimizer, modules and puts the networks on the right
        devices. Optionally loads checkpoint if already present.

        This function gets executed at the beginning of every training.
        """
        self.experiment_folder = os.path.join(
            self.hparams.output_folder, self.hparams.experiment_name
        )
        if self.hparams.debug:
            self.experiment_folder = "tmp_" + get_random_string()
            logger.info(f"Created temporary folder for debug {self.experiment_folder}.")

        save_dir = os.path.join(self.experiment_folder, "save")
        if os.path.exists(save_dir):
            if len(os.listdir(save_dir)) != 0:
                # select which checkpoint and load it.
                checkpoint, path = select_and_load_checkpoint(save_dir)
                self.opt = checkpoint["optimizer"]
                self.lr_sched = checkpoint["lr_scheduler"]
                self.start_epoch = checkpoint["epoch"] + 1

                self.load_modules(path)

                if self.accelerator.is_local_main_process:
                    self.checkpointer = Checkpointer(
                        checkpoint["key"],
                        mode=checkpoint["mode"],
                        checkpoint_path=self.experiment_folder,
                    )

                    logger.info(f"Loaded existing checkpoint from {path}.")
            else:
                self.opt, self.lr_sched = self.configure_optimizers()
                self.start_epoch = 0

                self.checkpointer = Checkpointer(
                    "val_loss", checkpoint_path=self.experiment_folder
                )
        else:
            os.makedirs(self.experiment_folder, exist_ok=True)

            self.opt, self.lr_sched = self.configure_optimizers()
            self.start_epoch = 0

            self.checkpointer = Checkpointer(
                "val_loss", checkpoint_path=self.experiment_folder
            )

        self.accelerator = Accelerator()
        self.device = self.accelerator.device
        self.modules.to(self.device)
        print("Set device to ", self.device)

        convert = [self.modules, self.opt, self.lr_sched] + list(self.datasets.values())
        accelerated = self.accelerator.prepare(convert)
        self.modules, self.opt, self.lr_sched = accelerated[:3]
        for i, key in enumerate(self.datasets):
            self.datasets[key] = accelerated[-(i + 1)]

    def on_train_end(self):
        """Runs at the end of each training. Cleans up before exiting."""
        if self.hparams.debug:
            logger.info(f"Removed temporary folder {self.experiment_folder}.")
            shutil.rmtree(self.experiment_folder)

        if self.accelerator.is_local_main_process:
            self.checkpointer.close()

    def train(
        self,
        epochs: int = 1,
        datasets: Dict = {},
        metrics: List[Metric] = [],
        debug: bool = False,
    ) -> None:
        """
        This method trains the model on the provided training dataset for the
        specified number of epochs. It tracks training metrics and can
        optionally perform validation during training, if the validation set is
        provided.

        Arguments
        ---------
        epochs : int
            The number of training epochs.
        datasets : Dict
            A dictionary of dataset loaders. Dataloader should be mapped to keys
            "train", "val", and "test".
        metrics : Optional[List[Metric]]
            A list of metrics to track during training. Default is an empty list.
        debug : bool
            Whether to run in debug mode. Default is False. If in debug mode,
            only runs for few epochs
            and with few batches.

        """
        self.datasets = datasets
        self.metrics = metrics
        assert "train" in self.datasets, "Training dataloader was not specified."
        assert epochs > 0, "You must specify at least one epoch."

        self.debug = debug

        self.on_train_start()

        if self.accelerator.is_local_main_process:
            logger.info(
                f"Starting from epoch {self.start_epoch}."
                + f" Training is scheduled for {epochs} epochs."
            )
        with self.accelerator.autocast():
            for e in range(self.start_epoch, epochs):
                pbar = tqdm(
                    self.datasets["train"],
                    unit="batches",
                    ascii=True,
                    dynamic_ncols=True,
                    disable=not self.accelerator.is_local_main_process,
                )
                loss_epoch = 0
                pbar.set_description(f"Running epoch {e + 1}/{epochs}")
                self.modules.train()
                for idx, batch in enumerate(pbar):
                    if isinstance(batch, list):
                        batch = [b.to(self.device) for b in batch]

                    self.opt.zero_grad()

                    model_out = self(batch)
                    loss = self.compute_loss(model_out, batch)

                    self.accelerator.backward(loss)
                    self.opt.step()

                    for m in self.metrics:
                        m(model_out, batch, Stage.train, self.device)

                    running_train = {
                        "train_" + m.name: m.reduce(Stage.train) for m in self.metrics
                    }

                    running_train.update({"train_loss": loss_epoch / (idx + 1)})

                    loss_epoch += loss.item()
                    pbar.set_postfix(**running_train)

                    if self.debug and idx > 10:
                        break

                pbar.close()

                train_metrics = {
                    "train_" + m.name: m.reduce(Stage.train, True) for m in self.metrics
                }
                train_metrics.update({"train_loss": loss_epoch / (idx + 1)})

                if "val" in datasets:
                    val_metrics = self.validate()
                    if self.accelerator.is_local_main_process:
                        self.checkpointer(
                            self,
                            e,
                            train_metrics,
                            val_metrics,
                            lambda x: self.accelerator.unwrap_model(x),
                        )
                else:
                    val_metrics = train_metrics.update(
                        {"val_loss": loss_epoch / (idx + 1)}
                    )

                if e >= 1 and self.debug:
                    break

        self.on_train_end()
        return None

    @torch.no_grad()
    def validate(self) -> Dict:
        """Runs the validation step."""
        assert "val" in self.datasets, "Validation dataloader was not specified."
        self.modules.eval()

        pbar = tqdm(
            self.datasets["val"],
            unit="batches",
            ascii=True,
            dynamic_ncols=True,
            disable=not self.accelerator.is_local_main_process,
        )
        loss_epoch = 0
        pbar.set_description("Validation...")
        with self.accelerator.autocast():
            for idx, batch in enumerate(pbar):
                if isinstance(batch, list):
                    batch = [b.to(self.device) for b in batch]

                self.opt.zero_grad()

                model_out = self(batch)
                loss = self.compute_loss(model_out, batch)
                for m in self.metrics:
                    m(model_out, batch, Stage.val, self.device)

                loss_epoch += loss.item()
                pbar.set_postfix(loss=loss_epoch / (idx + 1))

                if self.debug and idx > 10:
                    break

        val_metrics = {"val_" + m.name: m.reduce(Stage.val, True) for m in self.metrics}
        val_metrics.update({"val_loss": loss_epoch / (idx + 1)})

        pbar.close()

        return val_metrics

    @torch.no_grad()
    def test(self, datasets: Dict = {}) -> None:
        """Runs the test steps."""
        assert "test" in self.datasets, "Test dataloader was not specified."
        self.modules.eval()

        pbar = tqdm(
            self.datasets["test"],
            unit="batches",
            ascii=True,
            dynamic_ncols=True,
            disable=not self.accelerator.is_local_main_process,
        )
        loss_epoch = 0
        pbar.set_description("Testing...")
        with self.accelerator.autocast():
            for idx, batch in enumerate(pbar):
                if isinstance(batch, list):
                    batch = [b.to(self.device) for b in batch]
                self.opt.zero_grad()

                model_out = self(batch)
                loss = self.compute_loss(model_out, batch)
                for m in self.metrics:
                    m(model_out, batch, Stage.test, self.device)

                loss_epoch += loss.item()
                pbar.set_postfix(loss=loss_epoch / (idx + 1))

        pbar.close()

        test_metrics = {
            "test_" + m.name: m.reduce(Stage.test, True) for m in self.metrics
        }
        test_metrics.update({"test_loss": loss_epoch / (idx + 1)})
        s_out = (
            "Testing "
            + " - ".join([f"{k}: {v:.2f}" for k, v in test_metrics.items()])
            + "; "
        )

        logger.info(s_out)

        return None



"""
Code for PhiNets (https://doi.org/10.1145/3510832).

Authors:
    - Francesco Paissan, 2023
    - Alberto Ancilotto, 2023
    - Matteo Beltrami, 2023
    - Matteo Tremonti, 2023
"""
from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchinfo import summary
import torch.ao.nn.quantized as nnq


def _make_divisible(v, divisor=8, min_value=None):
    """
    This function is taken from the original tf repo. It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py

    It ensures that all layers have a channel number that is divisible by divisor.

    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


def correct_pad(input_shape, kernel_size):
    """Returns a tuple for zero-padding for 2D convolution with downsampling"""
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    if input_shape[0] is None:
        adjust = (1, 1)
    else:
        adjust = (1 - input_shape[0] % 2, 1 - input_shape[1] % 2)

    correct = (kernel_size[0] // 2, kernel_size[1] // 2)

    return (
        int(correct[1] - adjust[1]),
        int(correct[1]),
        int(correct[0] - adjust[0]),
        int(correct[0]),
    )


def preprocess_input(x, **kwargs):
    """Normalise channels between [-1, 1]"""

    return (x / 128.0) - 1


def get_xpansion_factor(t_zero, beta, block_id, num_blocks):
    """Compute expansion factor based on the formula from the paper"""
    return (t_zero * beta) * block_id / num_blocks + t_zero * (
        num_blocks - block_id
    ) / num_blocks


class ReLUMax(torch.nn.Module):
    def __init__(self, max):
        super(ReLUMax, self).__init__()
        self.max = max

    def forward(self, x):
        return torch.clamp(x, min=0, max=self.max)


class SEBlock(torch.nn.Module):
    """Implements squeeze-and-excitation block"""

    def __init__(self, in_channels, out_channels, h_swish=True):
        """Constructor of SEBlock

        Args:
            in_channels ([int]): [Input number of channels]
            out_channels ([int]): [Output number of channels]
            h_swish (bool, optional): [Whether to use the h_swish]. Defaults to True.
        """
        super(SEBlock, self).__init__()

        self.se_conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            padding=0,
            bias=False,
        )

        self.se_conv2 = nn.Conv2d(
            out_channels, in_channels, kernel_size=1, bias=False, padding=0
        )

        if h_swish:
            self.activation = nn.Hardswish(inplace=True)
        else:
            self.activation = ReLUMax(6)

        # It serves for the quantization.
        # The behavior remains equivalent for the unquantized models.
        self.mult = nnq.FloatFunctional()

    def forward(self, x):
        """Executes SE Block

        Args:
            x ([Tensor]): [input tensor]

        Returns:
            [Tensor]: [output of squeeze-and-excitation block]
        """

        inp = x
        x = F.adaptive_avg_pool2d(x, (1, 1))
        x = self.se_conv(x)
        x = self.activation(x)
        x = self.se_conv2(x)
        x = torch.sigmoid(x)

        return self.mult.mul(inp, x)  # Equivalent to ``torch.mul(a, b)``


class DepthwiseConv2d(torch.nn.Conv2d):
    def __init__(
        self,
        in_channels,
        depth_multiplier=1,
        kernel_size=3,
        stride=1,
        padding=0,
        dilation=1,
        bias=False,
        padding_mode="zeros",
    ):
        out_channels = in_channels * depth_multiplier
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
            padding_mode=padding_mode,
        )


class SeparableConv2d(torch.nn.Module):
    """Implements SeparableConv2d"""

    def __init__(
        self,
        in_channels,
        out_channels,
        activation=torch.nn.functional.relu,
        kernel_size=3,
        stride=1,
        padding=0,
        dilation=1,
        bias=True,
        padding_mode="zeros",
        depth_multiplier=1,
    ):
        """Constructor of SeparableConv2d

        Args:
            in_channels ([int]): [Input number of channels]
            out_channels ([int]): [Output number of channels]
            kernel_size (int, optional): [Kernel size]. Defaults to 3.
            stride (int, optional): [Stride for conv]. Defaults to 1.
            padding (int, optional): [Padding for conv]. Defaults to 0.
            dilation (int, optional): []. Defaults to 1.
            bias (bool, optional): []. Defaults to True.
            padding_mode (str, optional): []. Defaults to 'zeros'.
            depth_multiplier (int, optional): [Depth multiplier]. Defaults to 1.
        """
        super().__init__()

        self._layers = torch.nn.ModuleList()

        depthwise = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=3,
            stride=stride,
            padding=0,
            dilation=1,
            groups=in_channels,
            bias=bias,
            padding_mode=padding_mode,
        )

        spatialConv = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=dilation,
            # groups=in_channels,
            bias=bias,
            padding_mode=padding_mode,
        )

        bn = torch.nn.BatchNorm2d(out_channels, eps=1e-3, momentum=0.999)

        self._layers.append(depthwise)
        self._layers.append(spatialConv)
        self._layers.append(bn)
        self._layers.append(activation)

    def forward(self, x):
        """Executes SeparableConv2d block

        Args:
            x ([Tensor]): [Input tensor]

        Returns:
            [Tensor]: [Output of convolution]
        """
        for layer in self._layers:
            x = layer(x)

        return x


class PhiNetConvBlock(nn.Module):
    """Implements PhiNet's convolutional block"""

    def __init__(
        self,
        in_shape,
        expansion,
        stride,
        filters,
        has_se,
        block_id=None,
        res=True,
        h_swish=True,
        k_size=3,
        dp_rate=0.05,
        divisor=1,
    ):
        """Defines the structure of a PhiNet convolutional block.

        Arguments
        -------
        in_shape : tuple
            Input shape of the conv block.
        expansion : float
            Expansion coefficient for this convolutional block.
        stride: int
            Stride for the conv block.
        filters : int
            Output channels of the convolutional block.
        block_id : int
            ID of the convolutional block.
        has_se : bool
            Whether to include use Squeeze and Excite or not.
        res : bool
            Whether to use the residual connection or not.
        h_swish : bool
            Whether to use HSwish or not.
        k_size : int
            Kernel size for the depthwise convolution.

        """
        super(PhiNetConvBlock, self).__init__()

        self.param_count = 0

        self.skip_conn = False

        self._layers = torch.nn.ModuleList()
        in_channels = in_shape[0]

        # Define activation function
        if h_swish:
            activation = nn.Hardswish(inplace=True)
        else:
            activation = ReLUMax(6)

        if block_id:
            # Expand
            conv1 = nn.Conv2d(
                in_channels,
                _make_divisible(int(expansion * in_channels), divisor=divisor),
                kernel_size=1,
                padding=0,
                bias=False,
            )

            bn1 = nn.BatchNorm2d(
                _make_divisible(int(expansion * in_channels), divisor=divisor),
                eps=1e-3,
                momentum=0.999,
            )

            self._layers.append(conv1)
            self._layers.append(bn1)
            self._layers.append(activation)

        if stride == 2:
            padding = correct_pad([res, res], 3)

        self._layers.append(nn.Dropout2d(dp_rate))

        d_mul = 1
        in_channels_dw = (
            _make_divisible(int(expansion * in_channels), divisor=divisor)
            if block_id
            else in_channels
        )
        out_channels_dw = in_channels_dw * d_mul
        dw1 = DepthwiseConv2d(
            in_channels=in_channels_dw,
            depth_multiplier=d_mul,
            kernel_size=k_size,
            stride=stride,
            bias=False,
            padding=k_size // 2 if stride == 1 else (padding[1], padding[3]),
        )

        bn_dw1 = nn.BatchNorm2d(
            out_channels_dw,
            eps=1e-3,
            momentum=0.999,
        )

        # It is necessary to reinitialize the activation
        # for functions using Module.children() to work properly.
        # Module.children() does not return repeated layers.
        if h_swish:
            activation = nn.Hardswish(inplace=True)
        else:
            activation = ReLUMax(6)

        self._layers.append(dw1)
        self._layers.append(bn_dw1)
        self._layers.append(activation)

        if has_se:
            num_reduced_filters = _make_divisible(
                max(1, int(out_channels_dw / 6)), divisor=divisor
            )
            se_block = SEBlock(out_channels_dw, num_reduced_filters, h_swish=h_swish)
            self._layers.append(se_block)

        conv2 = nn.Conv2d(
            in_channels=out_channels_dw,
            out_channels=filters,
            kernel_size=1,
            padding=0,
            bias=False,
        )

        bn2 = nn.BatchNorm2d(
            filters,
            eps=1e-3,
            momentum=0.999,
        )

        self._layers.append(conv2)
        self._layers.append(bn2)

        if res and in_channels == filters and stride == 1:
            self.skip_conn = True
            # It serves for the quantization.
            # The behavior remains equivalent for the unquantized models.
            self.op = nnq.FloatFunctional()

    def forward(self, x):
        """Executes PhiNet convolutional block

        Arguments:
            x : torch.Tensor
                Input to the convolutional block.

        Returns:
            Ouput of the convolutional block : torch.Tensor
        """

        if self.skip_conn:
            inp = x

        for layer in self._layers:
            x = layer(x)

        if self.skip_conn:
            return self.op.add(x, inp)  # Equivalent to ``torch.add(a, b)``

        return x


class PhiNet(nn.Module):
    def get_complexity(self):
        """Returns MAC and number of parameters of initialized architecture.

        Returns
        -------
            Dictionary with complexity characterization of the network. : dict

        Example
        -------
        .. doctest::

            >>> from micromind.networks import PhiNet
            >>> model = PhiNet((3, 224, 224))
            >>> model.get_complexity()
            {'MAC': 9817670, 'params': 30917}
        """
        temp = summary(
            self, input_data=torch.zeros([1] + list(self.input_shape)), verbose=0
        )

        return {"MAC": temp.total_mult_adds, "params": temp.total_params}

    def get_MAC(self):
        """Returns number of MACs for this architecture.

        Returns
        -------
            Number of MAC for this network. : int

        Example
        -------
        .. doctest::

            >>> from micromind.networks import PhiNet
            >>> model = PhiNet((3, 224, 224))
            >>> model.get_MAC()
            9817670
        """
        return self.get_complexity()["MAC"]

    def get_params(self):
        """Returns number of params for this architecture.

        Returns
        -------
            Number of parameters for this network. : int

        Example
        -------
        .. doctest::

            >>> from micromind.networks import PhiNet
            >>> model = PhiNet((3, 224, 224))
            >>> model.get_params()
            30917
        """
        return self.get_complexity()["params"]

    def __init__(
        self,
        input_shape: List[int],
        num_layers: int = 7,  # num_layers
        alpha: float = 0.2,
        beta: float = 1.0,
        t_zero: float = 6,
        include_top: bool = False,
        num_classes: int = 10,
        compatibility: bool = False,
        downsampling_layers: List[int] = [5, 7],  # S2
        conv5_percent: float = 0.0,  # S2
        first_conv_stride: int = 2,  # S2
        residuals: bool = True,  # S2
        conv2d_input: bool = False,  # S2
        pool: bool = False,  # S2
        h_swish: bool = True,  # S1
        squeeze_excite: bool = True,  # S1
        divisor: int = 1,
    ) -> None:
        """This class implements the PhiNet architecture.

        Arguments
        -------
        input_shape : tuple
            Input resolution as (C, H, W).
        num_layers : int
            Number of convolutional blocks.
        alpha: float
            Width multiplier for PhiNet architecture.
        beta : float
            Shape factor of PhiNet.
        t_zero : float
            Base expansion factor for PhiNet.
        include_top : bool
            Whether to include classification head or not.
        num_classes : int
            Number of classes for the classification head.
        compatibility : bool
            `True` to maximise compatibility among embedded platforms (changes network).

        """
        super(PhiNet, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.t_zero = t_zero
        self.num_layers = num_layers
        self.num_classes = num_classes

        if compatibility:  # disables operations hard for some platforms
            h_swish = False
            squeeze_excite = False

        # this hyperparameters are hard-coded. Defined here as variables just so
        # you can play with them.
        first_conv_filters = 48
        b1_filters = 24
        b2_filters = 48

        if not isinstance(num_layers, int):
            num_layers = round(num_layers)

        assert len(input_shape) == 3, "Expected 3 elements list as input_shape."
        in_channels = input_shape[0]
        res = max(input_shape[1], input_shape[2])  # assumes squared input
        self.input_shape = input_shape

        self.classify = include_top
        self._layers = torch.nn.ModuleList()

        # Define self.activation function
        if h_swish:
            activation = nn.Hardswish(inplace=True)
        else:
            activation = ReLUMax(6)

        mp = nn.MaxPool2d((2, 2))

        if not conv2d_input:
            pad = nn.ZeroPad2d(
                padding=correct_pad([res, res], 3),
            )

            self._layers.append(pad)

            sep1 = SeparableConv2d(
                in_channels,
                _make_divisible(int(first_conv_filters * alpha), divisor=divisor),
                kernel_size=3,
                stride=(first_conv_stride, first_conv_stride),
                padding=0,
                bias=False,
                activation=activation,
            )

            self._layers.append(sep1)
            # self._layers.append(activation)

            block1 = PhiNetConvBlock(
                in_shape=(
                    _make_divisible(int(first_conv_filters * alpha), divisor=divisor),
                    res / first_conv_stride,
                    res / first_conv_stride,
                ),
                filters=_make_divisible(int(b1_filters * alpha), divisor=divisor),
                stride=1,
                expansion=1,
                has_se=False,
                res=residuals,
                h_swish=h_swish,
                divisor=divisor,
            )

            self._layers.append(block1)
        else:
            c1 = nn.Conv2d(
                in_channels, int(b1_filters * alpha), kernel_size=(3, 3), bias=False
            )

            bn_c1 = nn.BatchNorm2d(int(b1_filters * alpha))

            self._layers.append(c1)
            self._layers.append(activation)
            self._layers.append(bn_c1)

        block2 = PhiNetConvBlock(
            (
                _make_divisible(int(b1_filters * alpha), divisor=divisor),
                res / first_conv_stride,
                res / first_conv_stride,
            ),
            filters=_make_divisible(int(b1_filters * alpha), divisor=divisor),
            stride=2 if (not pool) else 1,
            expansion=get_xpansion_factor(t_zero, beta, 1, num_layers),
            block_id=1,
            has_se=squeeze_excite,
            res=residuals,
            h_swish=h_swish,
            divisor=divisor,
        )

        block3 = PhiNetConvBlock(
            (
                _make_divisible(int(b1_filters * alpha), divisor=divisor),
                res / first_conv_stride / 2,
                res / first_conv_stride / 2,
            ),
            filters=_make_divisible(int(b1_filters * alpha), divisor=divisor),
            stride=1,
            expansion=get_xpansion_factor(t_zero, beta, 2, num_layers),
            block_id=2,
            has_se=squeeze_excite,
            res=residuals,
            h_swish=h_swish,
            divisor=divisor,
        )

        block4 = PhiNetConvBlock(
            (
                _make_divisible(int(b1_filters * alpha), divisor=divisor),
                res / first_conv_stride / 2,
                res / first_conv_stride / 2,
            ),
            filters=_make_divisible(int(b2_filters * alpha), divisor=divisor),
            stride=2 if (not pool) else 1,
            expansion=get_xpansion_factor(t_zero, beta, 3, num_layers),
            block_id=3,
            has_se=squeeze_excite,
            res=residuals,
            h_swish=h_swish,
            divisor=divisor,
        )

        self._layers.append(block2)
        if pool:
            self._layers.append(mp)
        self._layers.append(block3)
        self._layers.append(block4)
        if pool:
            self._layers.append(mp)

        block_id = 4
        block_filters = b2_filters
        spatial_res = res / first_conv_stride / 4
        in_channels_next = _make_divisible(int(b2_filters * alpha), divisor=divisor)
        while num_layers >= block_id:
            if block_id in downsampling_layers:
                block_filters *= 2
                if pool:
                    self._layers.append(mp)

            pn_block = PhiNetConvBlock(
                (in_channels_next, spatial_res, spatial_res),
                filters=_make_divisible(int(block_filters * alpha), divisor=divisor),
                stride=(2 if (block_id in downsampling_layers) and (not pool) else 1),
                expansion=get_xpansion_factor(t_zero, beta, block_id, num_layers),
                block_id=block_id,
                has_se=squeeze_excite,
                res=residuals,
                h_swish=h_swish,
                k_size=(5 if (block_id / num_layers) > (1 - conv5_percent) else 3),
                divisor=divisor,
            )

            self._layers.append(pn_block)
            in_channels_next = _make_divisible(
                int(block_filters * alpha), divisor=divisor
            )
            spatial_res = (
                spatial_res / 2 if block_id in downsampling_layers else spatial_res
            )
            block_id += 1

        if include_top:
            # Includes classification head if required
            self.classifier = nn.Sequential(
                nn.AdaptiveAvgPool2d((1, 1)),
                nn.Flatten(),
                nn.Linear(
                    _make_divisible(int(block_filters * alpha), divisor=divisor),
                    num_classes,
                    bias=True,
                ),
            )

    def forward(self, x):
        """Executes PhiNet network

        Arguments
        -------
            x : torch.Tensor
                Network input.

        Returns
        ------
            Logits if `include_top=True`, otherwise embeddings : torch.Tensor
        """
        for layers in self._layers:
            x = layers(x)

        if self.classify:
            x = self.classifier(x)

        return x



"""
Conversion from pytorch to different standard formats
for inference (ONNX, OpenVINO, tflite).

Authors:
    - Francesco Paissan, 2023
    - Alberto Ancilotto, 2023
"""
from pathlib import Path
from loguru import logger
from typing import Union
import torch.nn as nn
import torch
import os


@torch.no_grad()
def convert_to_onnx(
    net: nn.Module,
    save_path: Union[Path, str] = "model.onnx",
    simplify: bool = False,
    replace_forward: bool = False,
):
    """Converts nn.Module to onnx and saves it to save_path.
    Optionally simplifies it."""
    save_path = Path(save_path)
    os.makedirs(save_path.parent, exist_ok=True)
    x = torch.zeros([1] + list(net.input_shape))

    if replace_forward:
        # add forward to ModuleDict
        bound_method = net.forward.__get__(net.modules, net.modules.__class__)
        setattr(net.modules, "forward", bound_method)

        net.modules.input_shape = net.input_shape
        net = net.modules
        x = [torch.zeros([1] + list(net.input_shape)), None]

    torch.onnx.export(
        net.cpu(),
        x,
        save_path,
        verbose=False,
        input_names=["input", "labels"],
        output_names=["output"],
        opset_version=11,
    )

    if simplify:
        import onnx
        import onnxsim

        onnx_model = onnx.load(save_path)
        onnx_model, check = onnxsim.simplify(onnx_model)
        onnx.save(onnx_model, save_path)

    logger.info(f"Saved converted ONNX model to {save_path}.")

    return save_path


@torch.no_grad()
def convert_to_openvino(
    net: nn.Module, save_path: Path, replace_forward: bool = False
) -> str:
    """Converts nn.Module to OpenVINO."""
    try:
        import os

        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        import sys
        from pathlib import Path
        from loguru import logger

        import onnx
        from onnx_tf.backend import prepare
        from openvino.tools.mo import main as mo_main

    except Exception as e:
        print(str(e))
        print("Did you install micromind with conversion capabilities?")
        print("Please try again after pip install micromind[conversion].")
        exit(0)
    os.makedirs(save_path, exist_ok=True)
    if not isinstance(save_path, Path):
        save_path = Path(save_path)

    onnx_path = save_path.joinpath("model.onnx")
    onnx_model = onnx.load(
        convert_to_onnx(net, onnx_path, simplify=True, replace_forward=replace_forward)
    )

    tf_rep = prepare(onnx_model)

    # Get the input tensor shape
    input_tensor = tf_rep.signatures[tf_rep.inputs[0]]
    input_shape = input_tensor.shape
    input_shape_str = "[" + ",".join([str(x) for x in input_shape]) + "]"

    cmd = [
        sys.executable,
        mo_main.__file__,
        "--input_model",
        str(onnx_path),
        "--input_shape",
        input_shape_str,
        "--output_dir",
        str(save_path),
        "--data_type",
        "FP32",
        "--silent",
        "True",
    ]

    os.popen(" ".join(cmd)).read()

    logger.info(f"Saved converted OpenVINO model to {save_path}.")

    return str(save_path.joinpath("model.xml"))


@torch.no_grad()
def convert_to_tflite(
    net: nn.Module,
    save_path: Path,
    batch_quant: torch.Tensor = None,
    replace_forward: bool = False,
) -> None:
    """Converts nn.Module to tf_lite, optionally quantizes it."""
    try:
        import os

        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        import shutil
        import sys
        from pathlib import Path
        from loguru import logger

        import numpy as np
        import tensorflow as tf

    except Exception as e:
        print(str(e))
        print("Did you install micromind with conversion capabilities?")
        print("Please try again after pip install micromind[conversion].")
        exit(0)
    if not isinstance(save_path, Path):
        save_path = Path(save_path)

    if not (batch_quant is None):
        batch_quant = batch_quant.cpu()

    vino_sub = save_path.joinpath("vino")
    os.makedirs(vino_sub, exist_ok=True)
    vino_path = convert_to_openvino(net, vino_sub, replace_forward=replace_forward)
    if os.name == "nt":
        openvino2tensorflow_exe_cmd = [
            sys.executable,
            os.path.join(
                os.path.dirname(sys.executable), "Scripts", "openvino2tensorflow"
            ),
        ]
    else:
        openvino2tensorflow_exe_cmd = ["openvino2tensorflow"]

    cmd = openvino2tensorflow_exe_cmd + [
        "--model_path",
        str(vino_path),
        "--model_output_path",
        str(save_path),
        "--output_saved_model",
        "--output_no_quant_float32_tflite",
        "--non_verbose",
    ]

    os.popen(" ".join(cmd)).read()

    shutil.rmtree(vino_sub)

    if not (batch_quant is None):
        converter = tf.lite.TFLiteConverter.from_saved_model(str(save_path))

        def representative_dataset():
            for i, sample in enumerate(batch_quant):
                yield [np.expand_dims(sample.cpu().numpy(), axis=0)]

        converter.optimizations = [tf.lite.Optimize.DEFAULT]
        converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]
        converter.inference_input_type = tf.float32
        converter.inference_output_type = tf.float32
        converter.representative_dataset = representative_dataset

        tflite_quant_model = converter.convert()

        with open(save_path.joinpath("model.int8.tflite"), "wb") as f:
            f.write(tflite_quant_model)

    logger.info(f"Saved converted TFLite model to {save_path}.")



from .core import MicroMind, Metric, Stage

# Package version
__version__ = "0.1.0"

"""datasets_info is a dictionary that contains information about the attributes
of the datasets.
This dictionary is used in networks.py inside the from_pretrained class method
in order to examine the inputs and initialize the PhiNet or, in case of
mismatching between dataset and Nclasses, raise an AssertionError."""
datasets_info = {
    "CIFAR-100": {"Nclasses": 100, "NChannels": 3, "ext": ".pth.tar"},
    "CIFAR-10": {"Nclasses": 10, "NChannels": 3, "ext": ".pth.tar"},
    "ImageNet-1k": {"Nclasses": 1000, "NChannels": 3, "ext": ".pth.tar"},
    "MNIST": {"Nclasses": 10, "NChannels": 1, "ext": ".pth.tar"},
}



import argparse


def parse_arguments():
    parser = argparse.ArgumentParser(description="General configuration for micromind.")

    parser.add_argument("--lr", type=float, default=0.001, help="Learning rate.")
    parser.add_argument(
        "--optimizer",
        dest="opt",
        default="adam",
        choices=["adam", "sgd"],
        help="Optimizer name.",
    )
    parser.add_argument(
        "--experiment_name", default="exp", help="Name of the experiment."
    )
    parser.add_argument(
        "--output_folder", default="results", help="Output folder path."
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Run in debug mode to check train and validation steps.",
    )

    args = parser.parse_args()
    return args



"""
micromind checkpointer. Unwraps models and saves the to disk with optimizer's
state etc.

Authors:
    - Francesco Paissan, 2023
"""
from typing import Union, Dict, Callable
from loguru import logger
from pathlib import Path
import os

import torch


class Checkpointer:
    def __init__(
        self,
        key: str,
        mode: str = "min",
        top_k: int = 5,
        checkpoint_path: Union[str, Path] = ".",
    ) -> None:
        assert mode in ["max", "min"], "Checkpointer mode can be only max or min."
        self.key = key
        self.mode = mode
        self.top_k = 5

        self.bests = [torch.inf] * self.top_k
        self.check_paths = [""] * self.top_k
        self.root_dir = checkpoint_path
        self.save_dir = os.path.join(self.root_dir, "save")
        os.makedirs(self.save_dir, exist_ok=True)
        self.fstream = open(os.path.join(self.root_dir, "train_log.txt"), "a")

    def __call__(
        self,
        mind,
        epoch: int,
        train_metrics: Dict,
        metrics: Dict,
        unwrap: Callable = lambda x: x,
    ) -> Union[Path, str]:
        s_out = (
            f"Epoch {epoch}: "
            + " - ".join([f"{k}: {v:.2f}" for k, v in train_metrics.items()])
            + "; "
        )
        s_out += " - ".join([f"{k2}: {v2:.4f}" for k2, v2 in metrics.items()]) + ".\n"
        self.fstream.write(s_out)
        logger.info(s_out)
        base_save = {
            "key": self.key,
            "mode": self.mode,
            "epoch": epoch,
            "optimizer": mind.opt,
            "lr_scheduler": mind.lr_sched,
        }
        to_remove = None
        if self.mode == "min":
            if metrics[self.key] <= min(self.bests):
                id_best = self.bests.index(min(self.bests))
                to_remove = self.check_paths[id_best]

                self.check_paths[id_best] = os.path.join(
                    self.save_dir,
                    f"epoch_{epoch}_{self.key}_{metrics[self.key]:.4f}.ckpt",
                )

                base_save.update(
                    {k: unwrap(v).state_dict() for k, v in mind.modules.items()}
                ),
                torch.save(base_save, self.check_paths[id_best])
        elif self.mode == "max":
            if metrics[self.key] >= max(self.bests):
                id_best = self.bests.index(min(self.bests))
                to_remove = self.check_paths[id_best]

                self.check_paths[id_best] = os.path.join(
                    self.save_dir,
                    f"epoch_{epoch}_{self.key}_{metrics[self.key]:.4f}.ckpt",
                )

                base_save.update(
                    {k: unwrap(v).state_dict() for k, v in mind.modules.items()}
                ),
                torch.save(base_save, self.check_paths[id_best])

        if to_remove is not None and to_remove != "":
            logger.info(f"Generated better checkpoint. Deleting {to_remove}.")
            os.remove(to_remove)

        if self.mode == "max":
            return self.check_paths[self.bests.index(max(self.bests))]
        elif self.mode == "min":
            return self.check_paths[self.bests.index(min(self.bests))]

    def close(self):
        self.fstream.close()



"""
micromind helper functions.

Authors:
    - Francesco Paissan, 2023
"""
from typing import Union, Dict, Tuple
from pathlib import Path
import random
import string
import torch
import os


def get_value_from_key(s: str, key: str, cast=float) -> float:
    dat = s.split(f"{key}_")[-1]

    if "ckpt" in dat:
        dat = dat.split(".ckpt")[0]

    return cast(dat)


def select_and_load_checkpoint(path: Union[Path, str]) -> Tuple[Dict, str]:
    checkpoints = os.listdir(path)
    checkpoints = [os.path.join(path, c) for c in checkpoints]

    dat = torch.load(checkpoints[0])
    selected_key, selected_mode = dat["key"], dat["mode"]

    values = [get_value_from_key(str(c), selected_key) for c in checkpoints]

    best_key = min(values) if selected_mode == "min" else max(values)
    best_checkpoint = checkpoints[values.index(best_key)]

    return torch.load(best_checkpoint), best_checkpoint


def get_random_string(length=10):
    letters = string.ascii_lowercase
    result_str = "".join(random.choice(letters) for i in range(length))

    return result_str
